(* ========================================================================= *)
(*                                                                           *)
(*                             SFG.ml                                        *)
(*                                                                           *)
(*        (c) Copyright, Sidi Mohamed BEILLAHI, Muhammad Umair Siddique      *)
(*                       2016                                                *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <beillahi@encs.concordia.ca>,                                *)
(*       	<muh_sidd@ece.concordia.ca>                                 *)
(*                                                                           *)
(* Last update: April 22, 2016                                               *)
(*                                                                           *)
(* ========================================================================= *)


(*-------------REQUIRED LIBRARIES ---------------------*)





needs "Multivariate/make_complex.ml";;



(*-----------------------------------------------------*)
(*----------------Formalization------------------------*)
(*-----------------------------------------------------*)
(*--------------------**********************------------------*)

(*-------------Generating Numbers----------------------*)
let numbers_1 = map (num_CONV o mk_small_numeral) (1--2000);;
let numbers_2 = (insert ADD1 (map (num_CONV o mk_small_numeral) (2--2000)));;
let NUMBER_SFG_TAC1 xs = REWRITE_TAC numbers_1
THEN REWRITE_TAC xs ;;
let NUMBER_SFG_TAC2 xs = REWRITE_TAC numbers_2
THEN REWRITE_TAC xs ;;
(*-----------------------------------------------------*)
let snd_of_trpl = 
	define ` snd_of_trpl (a,b,c) = b`;;
let fst_of_trpl = 
	define ` fst_of_trpl (a,b,c) = a`;;
let lst_of_trpl = 
	define ` lst_of_trpl (a,b,c) = c`;;

let snd_of_four = 
	define ` snd_of_four (a,b,c,d) = b`;;
let fst_of_four = 
	define ` fst_of_four (a,b,c,d) = a`;;
let thd_of_four = 
	define ` thd_of_four (a,b,c,d) = c`;;
let lst_of_four = 
	define ` lst_of_four (a,b,c,d) = d`;;

new_type_abbrev("branch",`:num#complex#num`);; 
new_type_abbrev("path",`:(branch)list`);; 
new_type_abbrev("SFG",`:(branch)list#num#num#num`);;
new_type_abbrev("KPH",`:num#((num)list)#(((num) list) list)`);; 


(*------------------------------------------------*)
let minus_one_aid = define 
		`(minus_one_aid ((k:num)+1) = k)`;;
let minus_one = define 
		`(minus_one (k:num) = if (k=1) 
				then 0 else (minus_one_aid k))`;;



(* verified if l1[k]=n so returning l2[k] and 
 continuing like that *)
let EC_G_creation_aid = define 
		`!(n:num) (l1: (num) list) (l2: (num) list). 
		(EC_G_creation_aid n (CONS e1 l1) (CONS e2 l2) = 
		if(e1=n) then (CONS e2 (EC_G_creation_aid n l1 l2)) 
		else  (EC_G_creation_aid n l1 l2)) /\ 
		(EC_G_creation_aid n [] [] = [])`;;
			
(* Create the list of list G *)
let EC_start_nodes = define 
		`!(l : path). 
		(EC_start_nodes (CONS e l) = 
	(CONS (fst_of_trpl e) (EC_start_nodes l))) /\ 
		(EC_start_nodes [] = []) `;;

let EC_end_nodes = define 
		`!(l: path). 
		(EC_end_nodes [] = []) /\
		(EC_end_nodes (CONS e l) = 
	(CONS (lst_of_trpl e) (EC_end_nodes l)))`;;
					  
let EC_G_creation = define 
 `!(k: num) (t1: (num) list) (t2: (num) list) (t3: (num) list).   
 (EC_G_creation k (CONS e1 t1) (CONS e2 t2) (CONS e3 t3)  = 
		(CONS (EC_G_creation_aid (k+1) (CONS e1 t1) (CONS e2 t2)) 
 (EC_G_creation (k+1) (CONS e1 t1) (CONS e2 t2) t3))) 
 /\ (EC_G_creation (k: num) (t1: (num) list) 
			(t2: (num) list) [] = []) `;;
			
			
			
					  
(* Creates a list of zeros *)
let EC_create_vector =  define 
			`!(n:num). 
			(EC_create_vector n = (REPLICATE n 0))`;;

(* Exchanges the first element of a list 
by another given element *)
let EC_replace_vector_head = define 
			`!(e:num) (l: (num) list). 
		(EC_replace_vector_head  [] e = []) /\
		(EC_replace_vector_head (CONS e1 l) e = (CONS e l))`;;
				
(* Creates a matrix m x m  of zeros *)
let EC_create_matrix = define 
			`!(m:num). 
			(EC_create_matrix m = (REPLICATE m (REPLICATE m 0)))`;;



(* Add one to the head of a given list *)

let EC_add_one_to_list_head = define 
	`!(l: (num) list). 
	(EC_add_one_to_list_head [] = []) /\
	(EC_add_one_to_list_head (CONS e l) = (CONS (e+1) l)) `;;

					 
(* put o in the case P[k+1] *)
let EC_remove_node = define 
	`!(P: (num) list) (k:num). 
	      (EC_remove_node (CONS (e:num) P) (k+1) = 
	       (CONS e (EC_remove_node P k))) /\ 
		(EC_remove_node (CONS e P)  0 = (CONS 0 P))`;;
					
(* Test if a given element exists in a given list *)
				
let NOT_IN_LIST = define 
	`!(e1 :num) (t: (num) list). 
	(NOT_IN_LIST e1 [] = T) /\ 
	(NOT_IN_LIST e1 (CONS e2 t) = 
	 if(e1=e2) then F else  
	if (e2=0) then T else T /\ (NOT_IN_LIST e1 t))`;;

		  
(* Modifies H as H[(k+1),the first j where (H[(k+1),j]=0)]=e1 *)
let EC_H_MODIFIE0 = define 
		`!(e2: num) (t: (num) list).  
		(EC_H_MODIFIE0 [] e2 = [e2]) /\
		(EC_H_MODIFIE0 (CONS e t) e2 = 
		    if(e = 0) then (CONS e2 t) 
		   else (CONS e (EC_H_MODIFIE0 t e2)))`;;
		   
let EC_H_MODIFIE1 = define 
		`!(H: ((num) list) list) (e1: num) (k:num).
		(EC_H_MODIFIE1 e1 [] k = []) /\
		(EC_H_MODIFIE1 e1 (CONS h H) (k+1) = 
		(CONS h (EC_H_MODIFIE1 e1 H k))) /\ 
		(EC_H_MODIFIE1 e1 (CONS h1 H) 0 = 
			(CONS (EC_H_MODIFIE0 h1 e1) H))`;;
				
let EC_H_MODIFIE2 = define 
		`!(H: ((num) list) list) (k:num) (n:num).
		(EC_H_MODIFIE2 [] k n = []) /\
		(EC_H_MODIFIE2 (CONS h H) (k+1) n = 
		(CONS h (EC_H_MODIFIE2 H k n))) /\ 
		(EC_H_MODIFIE2 (CONS h1 H) 0 n = (CONS (REPLICATE n 0) H))`;;
		
let EC_H1_REPLACE = define 
		`!(H: ((num) list) list) (k:num) (l:num list).
		(EC_H1_REPLACE l k [] = []) /\
		(EC_H1_REPLACE l (k+1) (CONS h H) = 
		(CONS h (EC_H1_REPLACE l k H))) /\ 
		(EC_H1_REPLACE l 0 (CONS h H) = (CONS l H))`;;
				
let EC_H_MODIFIE = define 
	`!(e1: num) (H: ((num) list) list) (k1:num) (k2:num). 
		(EC_H_MODIFIE e1 H k1 k2 = 
		(EC_H_MODIFIE2 (EC_H_MODIFIE1 e1 H k1) k2 (LENGTH H)))`;;
						  
let EC_H1_MODIFIE = define 
	`!(e1: num) (H1: ((num) list) list) (l:num list). 
		(EC_H1_MODIFIE e1 H1 [] = H1) /\
		(EC_H1_MODIFIE e1 H1 (CONS h l) = 
		(EC_H1_MODIFIE e1 (EC_H1_REPLACE (EC_H_MODIFIE0 (EL (minus_one h) H1) e1) (minus_one h) H1) l) )`;;
(* P[j (where j is the first for which P[j] =0 )] =e  *) 
let EC_add_node = define 
		`!(P : (num) list).
		(EC_add_node [] = []) /\ 
		(EC_add_node (CONS e P) = 
			if (e=0) then (EC_add_node P) 
			else (CONS e (EC_add_node P)))`;;
				   
				   
(* Test if l[0] = n *)
let EECTEST = define 
		`!(l: (num) list) (n:num).
		(EECTEST [] n = F) /\
		(EECTEST (CONS e l) n = if(e=n) then T else F)`;;

					  
(* Test if P[0] is in G[P[k]-1:] *)
let IN_LIST = define 
	`!(e:num) (t : (num) list). 
	(IN_LIST e [] = F) /\
	(IN_LIST e (CONS e1 t) = 
		if (e=e1) then T  
		else (F \/ (IN_LIST e t))) `;;

let EC_TEST = define 
		`!(P: (num) list) (k:num) (G: ((num) list) list).
EC_TEST P k G = (IN_LIST (EL 0 P) (EL (minus_one (EL k P)) G)) `;;
					   
(* Return the triple which represents the branch started 
from the node e1 and end in the node e2 *)
let EC_REWRITING2 = define 
		`!(e1:num) (e2:num) (l: path).
		(EC_REWRITING2 e1 e2 [] = Cx(&0)) /\
		(EC_REWRITING2 e1 e2 (CONS e l) = 
		if((e1= (fst_of_trpl e)) /\ (e2 = (lst_of_trpl e))) 
		then (snd_of_trpl e) 
		else (EC_REWRITING2 e1 e2 l) ) `;;
(*---------------------------------------------------------------------
***********************************************************************
----------------------------------------------------------------------*)					   
(* Extract (n+1) for which t1[n]>t2[m+1] and t1[n] 
not in l1 and t2[n] not in l2 *)	   
let EC_next_node = define 
	`!(n:num) (m:num) (t1: (num) list) (t2: (num) list) 
	(l1: (num) list) (l2: (num) list) (h1: (num) list). 
	(EC_next_node n m [] t2 l1 l2 h1 = 0) /\
	(EC_next_node n (m+1) (CONS e1 t1) (CONS e2 t2) l1 l2 h1 = 
	(EC_next_node n m (CONS e1 t1) t2 l1 l2 h1)) 
	/\ (EC_next_node n 0 (CONS e1 t1) (CONS e2 t2) l1 l2 h1 = 
	if ((NOT_IN_LIST e1 l1) /\ (NOT_IN_LIST e1 l2) /\ (NOT_IN_LIST e1 h1)) then (n+1)
	else (EC_next_node (n+1) 0 t1 (CONS e2 t2) l1 l2 h1)) `;;
					 
			
			
(* Exchange t[k+1] = e *)			
let EC_replace_first_zero = define 
		`!(k:num) (e:num) (t: (num) list).
		(EC_replace_first_zero k e [] = []) /\
		(EC_replace_first_zero (k+1) e (CONS e1 t) = 
		(CONS e1 (EC_replace_first_zero k e t))) /\ 
		(EC_replace_first_zero 0 e (CONS e1 t) = (CONS e t)) `;;
			
(*-------------------Feedback loop----------------------*)



let EC_next_node_main = define 
`!(k:num) (P: (num) list) (H: ((num) list) list) (H1: ((num) list) list) (G:((num) list) list).
		EC_next_node_main k P H H1 G = 
		(EC_next_node 0 k (EL (minus_one (EL k P)) G) P P 
			(EL (minus_one (EL k P)) H) (EL (minus_one (EL k P)) H1))`;;
			
(* Constructs a path starts from the node P[0] *)
let EC_search_path = define 
`!(k:num) (P: (num) list) (H: ((num) list) list) (H1: ((num) list) list) (G:((num) list) list) (j:num).
	(EC_search_path k P H H1 G j  = 
	if(~(j=0)) then (EC_search_path (k+1)
	(EC_replace_first_zero (k+1) (EL (minus_one j) (EL 
	(minus_one (EL k P)) G)) P) H H1 G 
	(EC_next_node_main (k+1) 
	(EC_replace_first_zero (k+1) (EL (minus_one j) (EL 
	(minus_one (EL k P)) G)) P) H H1 G)) 
	else (if(~(k=0)) then (k,P,(EC_H_MODIFIE (EL k P) H 
	(minus_one (EL (minus_one k) P)) (minus_one (EL k P)) )) else (k,P,H))) `;;

let EC_POINT_AID = define 
	`!(l: KPH) (H1: ((num) list) list) (G : ((num) list) list).
		(EC_POINT_AID l H1 G = 
		(EC_search_path (minus_one (fst_of_trpl l)) 
		(EC_remove_node (snd_of_trpl l) (fst_of_trpl l)) 
		(lst_of_trpl l) H1 G (EC_next_node_main 
		(minus_one (fst_of_trpl l)) 
		(EC_remove_node (snd_of_trpl l) (fst_of_trpl l)) 
		(lst_of_trpl l) H1 G)))`;;

let EC_POINT = define 
	`!(l :KPH) (H1: ((num) list) list) (G: ((num) list) list) (t3: (num) list).
	(EC_POINT l H1 G [] = []) /\
	(EC_POINT l H1 G (CONS e3 t3) = 
	if((fst_of_trpl l) = 0) 
	then (if(EC_TEST (snd_of_trpl l) (fst_of_trpl l) G)
	then [(EC_add_node (snd_of_trpl l))] else [])
	else (APPEND (if(EC_TEST (snd_of_trpl l) (fst_of_trpl l) G) 
	then [(EC_add_node (snd_of_trpl l))]  else []) 
	(EC_POINT (EC_POINT_AID l H1 G) H1 G t3))) `;;
	
	   
let EC_ALL_POINT = define 
	`!(k:num) (P: (num) list) (H: ((num) list) list) (H1: ((num) list) list) 
	(G: ((num) list) list) (G1: ((num) list) list) (n:num) (m:num) (t: (num) list).
	(EC_ALL_POINT k P H H1 G G1 n m [] = []) /\
	(EC_ALL_POINT k P H H1 G G1 n m (CONS e t) =
    if((EECTEST P (n+1))) then [] else 
	(APPEND (EC_POINT (EC_search_path k P H H1 G 
	(EC_next_node_main k P H H1 G)) H1 G (REPLICATE (m*m) 0)) 
	(EC_ALL_POINT k (EC_add_one_to_list_head P) H (EC_H1_MODIFIE (EL 0 P) H1 
	(EL (minus_one (EL 0 P)) G1)) G G1 n (minus_one m) t)))`;;
		  






let EC_MAIN_AID = define `! (l1: (num) list) (l2: (num) list) (n:num)
(H: ((num) list) list) (l3: (num) list) (l4: (num) list). 
	EC_MAIN_AID n l1 H l2 l3 l4 = 
	(EC_ALL_POINT 0 l1 H H (EC_G_creation 0 l2 l3 l4) (EC_G_creation 0 l3 l2 l4) n n l4)`;; 
 
let EC_MAIN = define `!(l: path) (n:num). 
	EC_MAIN l n = 
	    (EC_MAIN_AID n (EC_replace_vector_head (EC_create_vector n) 1)  (EC_create_matrix n) 
	    (EC_start_nodes l) (EC_end_nodes l) (EC_create_vector n))`;; 

(* Return the path in the form of list of branches *)
let EC_REWRITING1 = define 
	`!(e2 : num) (t: (num) list) (l: path) (e:num).
	(EC_REWRITING1 e2 [] l = []) 
	  /\ (EC_REWRITING1 e2 [e] l = 
	  ([(EC_REWRITING2 e e2 l)])) 
	    /\ (EC_REWRITING1 e2 (CONS e1 (CONS e t)) l = 
	   (CONS (EC_REWRITING2 e1 e l) 
	   (EC_REWRITING1 e2 (CONS e t) l)))`;;



(* Return all the elementary circuits in 
the form of list of list of branches *)
let EC_REWRITING0 = define 
	`!(t: (num) list) (l: (complex) list) (e2:num).
	 (EC_REWRITING0 [] [] e2 = []) /\
	(EC_REWRITING0 (CONS e t) (CONS e1 l) e2 = 
	 if (t=[]) then [(e,e1,e2)] else (CONS (e,e1,(HD t)) 
	  (EC_REWRITING0 t l e2)) )`;;

let EC_REWRITING = define 
	`!(p: ((num) list) list) (l: path).
	(EC_REWRITING [] l = []) /\
	(EC_REWRITING (CONS t p) l = 
	(CONS (EC_REWRITING0 t (EC_REWRITING1 (HD t) t l) (HD t)) 
	(EC_REWRITING p l)))`;;

(* The main function which returns all elementary circuits *)
let EC = define `!(l: path) (n:num).
	EC l n = if(l=[]) then [] else 
       	 (EC_REWRITING (EC_MAIN l n) l) ` ;;
(*************************************************************)

(*-------------------Forward Path--------------------------*)
let FC_next_node = define 
	`!(n:num) (m:num) (t1: (num) list) (t2: (num) list) 
	(l1: (num) list) (l2: (num) list). 
	(FC_next_node n m [] t2 l1 l2 = 0) /\
	(FC_next_node n (m+1) (CONS e1 t1) (CONS e2 t2) l1 l2 = 
	(FC_next_node n m (CONS e1 t1) t2 l1 l2)) 
	/\ (FC_next_node n 0 (CONS e1 t1) (CONS e2 t2) l1 l2 = 
	if ((NOT_IN_LIST e1 l1) /\ (NOT_IN_LIST e1 l2)) then (n+1)
	else (FC_next_node (n+1) 0 t1 (CONS e2 t2) l1 l2)) `;;
					 

let FC_next_node_main = define 
	`!(k:num) (P: (num) list) (H: ((num) list) list) (G:((num) list) list).
	FC_next_node_main k P H G  = 
	(FC_next_node 0 0 (EL (minus_one (EL k P)) G) P P 
	(EL (minus_one (EL k P)) H))`;;
	  
	  
(* Constructs a path starts from the node P[0] *)
let FC_search_path = define 
`!(k:num) (P: (num) list) (H: ((num) list) list) (G:((num) list) list) (j:num).
	(FC_search_path k P H G j = if(~(j=0)) then (FC_search_path (k+1) 
	(EC_replace_first_zero (k+1) (EL (minus_one j) (EL 
	(minus_one (EL k P)) G)) P) H G (FC_next_node_main (k+1) 
	(EC_replace_first_zero (k+1) (EL (minus_one j) (EL 
	(minus_one (EL k P)) G)) P) H G)) else (if(~(k=0)) 
	then (k,P,(EC_H_MODIFIE (EL k P) H 
	(minus_one (EL (minus_one k) P)) (minus_one (EL k P)))) else (k,P,H))) `;;

let FC_POINT_AID = define 
	`!(l: KPH) (G : ((num) list) list).
	(FC_POINT_AID l G = (FC_search_path (minus_one (fst_of_trpl l)) 
	(EC_remove_node (snd_of_trpl l) (fst_of_trpl l)) 
  (lst_of_trpl l) G (FC_next_node_main (minus_one (fst_of_trpl l)) 
	(EC_remove_node (snd_of_trpl l) (fst_of_trpl l)) 
			(lst_of_trpl l) G)))`;;

(* Testing if two elements are the same *)
let FCTEST1 = define 
	`!(e: num)  (e1: num). (FCTEST1 e e1 = if (e=e1) then T else F)`;;

(* Tests if P[k] = stop *)
let FCTEST = define 
	`!(P: (num) list) (k:num) (stop: num).
	FCTEST P k stop  = (FCTEST1 (EL k P) (stop)) `;;

(* Calculates all forwards paths *)



let FC_FORWARD = define 
`!(l :KPH) (G: ((num) list) list) (t: (num) list) (stop: num).
	(FC_FORWARD l G [] stop = []) /\
	(FC_FORWARD l G (CONS e t) stop = if(((fst_of_trpl l) = 0) /\ 
	(~(FCTEST (snd_of_trpl l) (fst_of_trpl l) stop))) 
	then [] else (APPEND (if(FCTEST (snd_of_trpl l) (fst_of_trpl l) stop) 
	then [(EC_add_node (snd_of_trpl l))]  else []) 
	(FC_FORWARD (FC_POINT_AID l G) G t stop)))`;;

let FORWARD_CIRCUIT = define 
	`(!(k:num) (P: (num) list) (H: ((num) list) list) 
      (G: ((num) list) list) (stop:num) (n:num).
	  FORWARD_CIRCUIT k P H G stop n  =
	  (FC_FORWARD (FC_search_path k P H G 
	  (FC_next_node_main k P H G)) G 
	  (REPLICATE (n*n) 0) stop))`;;
	  
	  
	  (* Extracting the forwards paths in list of list *)
let FC_MAIN = define 
	`!(l: path) (n:num) (stop: num) (start: num).
	FC_MAIN l n stop start  = 
	(FORWARD_CIRCUIT 0 (EC_replace_vector_head (EC_create_vector n) start) 
	(EC_create_matrix n) (EC_G_creation 0 
	(EC_start_nodes l)                
	(EC_end_nodes l) 
	(EC_create_vector n)) stop n)`;; 
	

(* Return the path in the form of list of branches *)
let FC_REWRITING0 = define 
	`!(e:num) (t: (num) list) (l: path).
	(FC_REWRITING0 [] l = []) /\
	(FC_REWRITING0 [e] (l: path) = []) 
	 /\ (FC_REWRITING0 (CONS e1 (CONS e t)) l = 
	   (CONS (EC_REWRITING2 e1 e l) 
	   (FC_REWRITING0 (CONS e t) l)))`;;
	   
(* Return all the forward paths in the 
form of list of list of branches *)
let FC_REWRITING1 = define 
 `!(t: (num) list) (l: (complex) list).
 (FC_REWRITING1 t [] = []) /\
	(FC_REWRITING1 (CONS e t) (CONS e1 l) = 
	     if (t=[]) then [] else (CONS (e,e1,(HD t)) 
		 (FC_REWRITING1 t l)) )`;;

let FC_REWRITING = define 
	`!(p: ((num) list) list) (l: path).
	(FC_REWRITING [] l = []) /\
	(FC_REWRITING (CONS t p) l = 
	   (CONS (FC_REWRITING1 t (FC_REWRITING0 t l)) 
	   (FC_REWRITING p l)))`;;

(* The main function which returns all forward paths *)
let FC = define 
	`!(l: path) (n:num) (stop: num) (start: num).
		FC l n stop start = if(l=[]) then [] else 
			   (FC_REWRITING (FC_MAIN l n stop start) l) ` ;;
			   
(*-------------------Valid path and loop------------------------------*)




	
let SFG_INVERSE = define 
    `!(l: path).
	SFG_INVERSE [] = [] /\
	SFG_INVERSE (CONS e l) = 
	(CONS (lst_of_trpl e,snd_of_trpl e,fst_of_trpl e) (SFG_INVERSE l))`;;
	
	let SFG_TRANSPOSE = define 
    `!(sfg: SFG).
	SFG_TRANSPOSE sfg = (SFG_INVERSE (fst_of_four sfg), 
	(snd_of_four sfg),(lst_of_four sfg),(thd_of_four sfg)) `;;
	
let IS_VALID_PATH = define 
		`!(l : path). IS_VALID_PATH (CONS e l) =  
		if (l=[]) then T else (if((lst_of_trpl e) = 
		(fst_of_trpl (HD l))) 
		then (T /\ (IS_VALID_PATH l)) else F)`;;
	 
			
let IS_VALID_LOOP_AID1  = define
		`!(l: (num) list). (IS_VALID_LOOP_AID1 [] = T) /\ 
		(IS_VALID_LOOP_AID1 (CONS e l) = ((NOT_IN_LIST e l) /\ 
		(IS_VALID_LOOP_AID1 l)))`;;

let IS_VALID_LOOP_AID2  = define
		`!(l: path) (e:num). (IS_VALID_LOOP_AID2 e [] = F) /\ 
		(IS_VALID_LOOP_AID2 e (CONS e1 l) = 
		if (l= []) then (if (lst_of_trpl e1 =e) then T else F) 
		else ((lst_of_trpl e1 = fst_of_trpl (HD l)) /\ 
		(IS_VALID_LOOP_AID2 e l))) `;;
		
let IS_VALID_LOOP =  define 
	`!(l:path). IS_VALID_LOOP l = if (l =[]) then T
	else ((IS_VALID_LOOP_AID1 (EC_start_nodes l)) /\ 
	(IS_VALID_LOOP_AID2 (fst_of_trpl (HD l)) l))`;;
(*--------------------------------------------------------------------*)
(*-----------------*************Tactics*************-----------------*)
let component_of_trpl = 
	CONJ (CONJ fst_of_trpl snd_of_trpl) lst_of_trpl ;;
let component_of_four = 
	CONJ (CONJ (CONJ fst_of_four snd_of_four) thd_of_four) lst_of_four ;;
let minus = CONJ minus_one minus_one_aid ;;

let TAC_EC_G_creation_aid = NUMBER_SFG_TAC1[] 
THEN ONCE_REWRITE_TAC [minus; EC_G_creation_aid] 
THEN NUMBER_SFG_TAC2 [] THEN CONV_TAC NUM_REDUCE_CONV ;;

let TAC_EC_G_creation =  (REPEAT (CHANGED_TAC TAC_EC_G_creation_aid)) 
THEN NUMBER_SFG_TAC1[] THEN ONCE_REWRITE_TAC [minus; EC_G_creation] 
THEN NUMBER_SFG_TAC2 [] THEN CONV_TAC NUM_REDUCE_CONV ;;
(*-------------------------------------------------------------------*)
(*---------------------------Forward Tactics-------------------------*)
(*-------------------------------------------------------------------*)

let SIMPLIFIED_FH_TAC0 xs =  
 
 REWRITE_TAC xs THEN  
 NUMBER_SFG_TAC1 [EL;minus;HD;TL; EC_H_MODIFIE1; 
        LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE;REPLICATE] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_H_MODIFIE1; 
        LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
        EC_H_MODIFIE0;REPLICATE] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
        EC_H_MODIFIE0;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [FC_FORWARD] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;FCTEST;component_of_trpl
;FCTEST1] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;FCTEST;component_of_trpl
;FCTEST1] THEN  
CONV_TAC NUM_REDUCE_CONV THEN NUMBER_SFG_TAC1 [EL;minus; 
HD;TL;FCTEST;component_of_trpl;FCTEST1] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;FCTEST;component_of_trpl
;FCTEST1] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [FC_POINT_AID;component_of_trpl] THEN  
ONCE_REWRITE_TAC [FC_POINT_AID;component_of_trpl] THEN  
ONCE_REWRITE_TAC [FC_POINT_AID;component_of_trpl] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;component_of_trpl
;EC_remove_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;component_of_trpl
;EC_remove_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [FC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
FC_next_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
FC_next_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [FC_search_path] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_replace_first_zero] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_replace_first_zero] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [FC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
FC_next_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
FC_next_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV ;; 
 
 
let SIMPLIFIED_FC_TAC0 xs =   
 
REWRITE_TAC xs THEN ONCE_REWRITE_TAC [FC_search_path] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_replace_first_zero] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_replace_first_zero] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [FC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
FC_next_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
FC_next_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV ;; 
 
(*--------------------------------------------------------------*) 
(*------------------------Feedback Tactics----------------------*) 
(*--------------------------------------------------------------*) 
 
let SIMPLIFIED_H_TAC0 xs =  
 
REWRITE_TAC xs THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EC_H_MODIFIE1
;EC_H1_REPLACE;EC_H1_MODIFIE; 
            LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_H_MODIFIE1
;EC_H1_REPLACE;EC_H1_MODIFIE; 
            LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EC_H_MODIFIE0] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_H_MODIFIE0] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [EC_POINT] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EC_TEST; 
component_of_trpl;IN_LIST; 
REPLICATE] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_TEST; 
component_of_trpl;IN_LIST; 
REPLICATE] THEN CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EC_TEST; 
component_of_trpl;IN_LIST] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_TEST; 
component_of_trpl;IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [EC_POINT_AID;component_of_trpl] THEN  
ONCE_REWRITE_TAC [EC_POINT_AID;component_of_trpl] THEN  
ONCE_REWRITE_TAC [EC_POINT_AID;component_of_trpl] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;component_of_trpl
;EC_remove_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;component_of_trpl
;EC_remove_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [EC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_next_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_next_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [EC_POINT_AID;component_of_trpl] THEN  
ONCE_REWRITE_TAC [EC_POINT_AID;component_of_trpl] THEN  
CONV_TAC NUM_REDUCE_CONV  THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EC_TEST; 
component_of_trpl;IN_LIST; 
REPLICATE] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_TEST; 
component_of_trpl;IN_LIST; 
REPLICATE] THEN CONV_TAC NUM_REDUCE_CONV;; 
 
 
 
let SIMPLIFIED_H_TAC1 xs =  
 
REWRITE_TAC xs THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EC_H_MODIFIE1
;EC_H1_REPLACE;EC_H1_MODIFIE;
            LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_H_MODIFIE1
;EC_H1_REPLACE;EC_H1_MODIFIE; 
            LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EC_H_MODIFIE0] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_H_MODIFIE0] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [EC_POINT] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EC_TEST;component_of_trpl
;IN_LIST;REPLICATE] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_TEST;component_of_trpl
;IN_LIST;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EC_TEST;component_of_trpl
;IN_LIST] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EC_TEST;component_of_trpl
;IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [EC_POINT_AID;component_of_trpl] THEN  
ONCE_REWRITE_TAC [EC_POINT_AID;component_of_trpl] THEN  
ONCE_REWRITE_TAC [EC_POINT_AID;component_of_trpl] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;component_of_trpl
;EC_remove_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;component_of_trpl
;EC_remove_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN 
 REWRITE_TAC [EC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_next_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_next_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [EC_search_path] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_replace_first_zero] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_replace_first_zero] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [EC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_next_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_next_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV ;; 
 
 
 
let SIMPLIFIED_EC_TAC0 xs = 
 
NUMBER_SFG_TAC1 [LENGTH;EL;minus;HD;TL;EC_H1_REPLACE;EC_H1_MODIFIE] THEN  
NUMBER_SFG_TAC2 [LENGTH;EL;minus;HD;TL;EC_H1_REPLACE;EC_H1_MODIFIE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN 
NUMBER_SFG_TAC1 [LENGTH;EL;minus;HD;TL;EC_H_MODIFIE0] THEN  
NUMBER_SFG_TAC2 [LENGTH;EL;minus;HD;TL;EC_H_MODIFIE0] THEN  
CONV_TAC NUM_REDUCE_CONV THEN
REWRITE_TAC xs THEN ONCE_REWRITE_TAC [EC_search_path] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_replace_first_zero] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_replace_first_zero] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [EC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_next_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_next_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV ;; 
 
 
let SIMPLIFIED_ALL_TAC0 xs = 
 
REWRITE_TAC xs THEN REWRITE_TAC [EC_add_one_to_list_head]  
THEN CONV_TAC NUM_REDUCE_CONV THEN  
ONCE_REWRITE_TAC [EC_ALL_POINT] THEN  
REWRITE_TAC [EC_next_node_main] THEN  
NUMBER_SFG_TAC1 [LENGTH;EL;minus;HD;TL;EC_H1_REPLACE;EC_H1_MODIFIE;EC_H_MODIFIE0] THEN  
NUMBER_SFG_TAC2 [LENGTH;EL;minus;HD;TL;EC_H1_REPLACE;EC_H1_MODIFIE;EC_H_MODIFIE0] THEN  
CONV_TAC NUM_REDUCE_CONV THEN 
NUMBER_SFG_TAC1 [LENGTH;EL;minus;HD;TL;EC_H_MODIFIE0] THEN  
NUMBER_SFG_TAC2 [LENGTH;EL;minus;HD;TL;EC_H_MODIFIE0] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;IN_LIST] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_next_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_next_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST]  
THEN CONV_TAC NUM_REDUCE_CONV;; 


(*----------------------------------------------------------------------*)
let TAC_EC = (CHANGED_TAC (SIMPLIFIED_EC_TAC0 []) ORELSE (SIMPLIFIED_H_TAC0 [])) 
THEN (CHANGED_TAC (SIMPLIFIED_ALL_TAC0 []) ORELSE (SIMPLIFIED_EC_TAC0 []));;

let TAC_FC = (CHANGED_TAC (SIMPLIFIED_FC_TAC0 []) ORELSE (SIMPLIFIED_FH_TAC0 []));;

(*----------------------------------------------------------------------*)